import 'dart:async';

import 'package:analyzer/dart/analysis/context_builder.dart';
import 'package:analyzer/dart/analysis/context_locator.dart';
import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer/file_system/file_system.dart';
import 'package:analyzer/src/dart/analysis/driver.dart'
    show AnalysisDriver, AnalysisDriverGeneric;
import 'package:analyzer/src/dart/analysis/driver_based_analysis_context.dart';
import 'package:analyzer_plugin/plugin/plugin.dart';
import 'package:analyzer_plugin/protocol/protocol_common.dart' as plugin;
import 'package:analyzer_plugin/protocol/protocol_generated.dart' as plugin;
import 'package:logging/logging.dart';
import 'package:pub_semver/pub_semver.dart';

import 'analysis_options.dart';
import 'checker.dart';
import 'logger/log.dart';

final _logger = Logger('pt_intl.plugin');

class PTIntlPlugin extends ServerPlugin {
  PTIntlPlugin(ResourceProvider provider) : super(provider);

  @override
  List<String> get fileGlobsToAnalyze => <String>['**/*.dart'];

  @override
  String get name => 'pt_linter';

  @override
  String get version => '0.2.1';

  @override
  bool isCompatibleWith(Version serverVersion) => true;


  var _filesFromSetPriorityFilesRequest = <String>[];

  final Map<AnalysisDriverGeneric, AnalysisOptions> _options = {};

  @override
  AnalysisDriverGeneric createAnalysisDriver(plugin.ContextRoot contextRoot) {
    final rootPath = contextRoot.root;
    final locator =
        ContextLocator(resourceProvider: resourceProvider).locateRoots(
      includedPaths: <String>[rootPath],
      excludedPaths: <String>[
        ...contextRoot.exclude,
      ],
      optionsFile: contextRoot.optionsFile,
    );
    if (locator.isEmpty) {
      final error = StateError('Unexpected empty context');
      channel.sendNotification(plugin.PluginErrorParams(
        true,
        error.message,
        error.stackTrace.toString(),
      ).toNotification());

      throw error;
    }

    final builder = ContextBuilder(
      resourceProvider: resourceProvider,
    );

    final analysisContext = builder.createContext(contextRoot: locator.first);
    // ignore: avoid_as
    final context = analysisContext as DriverBasedAnalysisContext;
    final dartDriver = context.driver;
    try {
      final analysisOptions =
          (_options[dartDriver] ??= _getAnalysisOptions(dartDriver));

      runZonedGuarded(
        () {
          dartDriver.results.listen((analysisResult) {
            if (analysisResult is ResolvedUnitResult) {
              _processResult(
                dartDriver,
                analysisOptions,
                analysisResult,
              );
            } else if (analysisResult is ErrorsResult) {
              _logger
                  .severe('analysis produce errors. ${analysisResult.errors}');
            }
          });
        },
        (Object e, StackTrace stackTrace) {
          channel.sendNotification(
            plugin.PluginErrorParams(
              false,
              'pt_linter. Unexpected error: ${e.toString()}',
              stackTrace.toString(),
            ).toNotification(),
          );
        },
      );
    } catch (e, stackTrace) {
      channel.sendNotification(
        plugin.PluginErrorParams(
          false,
          'pt_linter. Unexpected error: ${e.toString()}',
          stackTrace.toString(),
        ).toNotification(),
      );
    }

    return dartDriver;
  }

  @override
  Future<plugin.EditGetFixesResult> handleEditGetFixes(
      plugin.EditGetFixesParams parameters) async {
    try {
      final driver = driverForPath(parameters.file) as AnalysisDriver;
      final analysisResult = await driver.getResult2(parameters.file);

      if (analysisResult is! ResolvedUnitResult) {
        return plugin.EditGetFixesResult([]);
      }

      final fixes = PTChecker(options: _options).start(
              driver, analysisResult.path, analysisResult.unit, analysisResult)
          .where((fix) =>
              fix.error.location.file == parameters.file &&
              fix.error.location.offset <= parameters.offset &&
              parameters.offset <=
                  fix.error.location.offset + fix.error.location.length &&
              fix.fixes.isNotEmpty)
          .toList();

      return plugin.EditGetFixesResult(fixes);
    } on Exception catch (e, stackTrace) {
      channel.sendNotification(
        plugin.PluginErrorParams(false, e.toString(), stackTrace.toString())
            .toNotification(),
      );

      return plugin.EditGetFixesResult([]);
    }
  }

  void _processResult(
    AnalysisDriver dartDriver,
    AnalysisOptions analysisOptions,
    ResolvedUnitResult analysisResult,
  ) {
    final path = analysisResult.path;

    final unit = analysisResult.unit;
    final isAnalyzed =
        dartDriver.analysisContext?.contextRoot.isAnalyzed(path) ?? false;
    final isExcluded = !isAnalyzed || analysisOptions.isExcluded(path);
    if (isExcluded) {
      _logger.finer('is not analyzed: $path '
          '(analyzed: $isAnalyzed / isExcluded: $isExcluded)');

      channel.sendNotification(
        plugin.AnalysisErrorsParams(
          path,
          <plugin.AnalysisError>[],
        ).toNotification(),
      );
      return;
    }

    try {
      final errors = PTChecker(options: _options).start(dartDriver, path, unit, analysisResult);

      channel.sendNotification(
        plugin.AnalysisErrorsParams(
          path,
          errors.map((e) => e.error).toList(),
        ).toNotification(),
      );
    } catch (e, stackTrace) {
      _logger.severe('Error while analysing file', e, stackTrace);
      channel.sendNotification(
        plugin.PluginErrorParams(
          false,
          e.toString(),
          stackTrace.toString(),
        ).toNotification(),
      );
    }
  }



  AnalysisOptions _getAnalysisOptions(AnalysisDriver analysisDriver) {
    final optionsPath = analysisDriver.analysisContext?.contextRoot.optionsFile;
    final exists = optionsPath?.exists ?? false;
    if (!exists || optionsPath == null) {
      _logger.warning('Unable to resolve optionsFile.');
      return AnalysisOptions(excludeGlobs: []);
    }
    return AnalysisOptions.loadFromYaml(optionsPath.readAsStringSync());
  }

  // @override
  // void sendNotificationsForSubscriptions(
  //     Map<String, List<AnalysisService>> subscriptions) {
  //   TODO: implement sendNotificationsForSubscriptions
  // }

  @override
  void contentChanged(String path) {
    super.driverForPath(path)?.addFile(path);
  }

  @override
  Future<plugin.AnalysisSetContextRootsResult> handleAnalysisSetContextRoots(
    plugin.AnalysisSetContextRootsParams parameters,
  ) async {
    final result = await super.handleAnalysisSetContextRoots(parameters);
    // The super-call adds files to the driver, so we need to prioritize them so they get analyzed.
    _updatePriorityFiles();

    return result;
  }

  @override
  Future<plugin.AnalysisSetPriorityFilesResult> handleAnalysisSetPriorityFiles(
    plugin.AnalysisSetPriorityFilesParams parameters,
  ) async {
    _filesFromSetPriorityFilesRequest = parameters.files;
    _updatePriorityFiles();

    return plugin.AnalysisSetPriorityFilesResult();
  }

  // https://github.com/dart-code-checker/dart-code-metrics/blob/e8e14d44b940a5b29d33a782432f853ee42ac7a0/lib/src/analyzer_plugin/analyzer_plugin.dart#L274
  /// AnalysisDriver doesn't fully resolve files that are added via `addFile`; they need to be either explicitly requested
  /// via `getResult`/etc, or added to `priorityFiles`.
  ///
  /// This method updates `priorityFiles` on the driver to include:
  ///
  /// - Any files prioritized by the analysis server via [handleAnalysisSetPriorityFiles]
  /// - All other files the driver has been told to analyze via addFile (in [ServerPlugin.handleAnalysisSetContextRoots])
  ///
  /// As a result, [_processResult] will get called with resolved units, and thus all of our diagnostics
  /// will get run on all files in the repo instead of only the currently open/edited ones!
  void _updatePriorityFiles() {
    final filesToFullyResolve = {
      // Ensure these go first, since they're actually considered priority; ...
      ..._filesFromSetPriorityFilesRequest,

      // ... all other files need to be analyzed, but don't trump priority
      for (final driver2 in driverMap.values)
        ...(driver2 as AnalysisDriver).addedFiles,
    };

    // From ServerPlugin.handleAnalysisSetPriorityFiles
    final filesByDriver = <AnalysisDriverGeneric, List<String>>{};
    for (final file in filesToFullyResolve) {
      final contextRoot = contextRootContaining(file);
      if (contextRoot != null) {
        // TODO(dkrutskikh): Which driver should we use if there is no context root?
        final driver = driverMap[contextRoot];
        if (driver != null) {
          filesByDriver.putIfAbsent(driver, () => <String>[]).add(file);
        }
      }
    }
    filesByDriver.forEach((driver, files) {
      driver.priorityFiles = files;
    });
  }
}

