import 'package:analyzer/dart/analysis/analysis_context.dart';
import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/src/dart/ast/constant_evaluator.dart';
import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/ast/visitor.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/source/line_info.dart';
import 'package:logging/logging.dart';
import 'package:path/path.dart' as path;
import 'package:source_gen/source_gen.dart';
import 'package:string_literal_finder_annotations/string_literal_finder_annotations.dart';
import 'exclude_path_checker.dart';

final _logger = Logger('pt_linter');

/// The main finder class which will use dart analyzer to analyse all
/// dart files in the given [basePath] and look for string literals.
/// Some literals will be (smartly) ignored which should not be localized.
class RtlWidgetFinder {
  RtlWidgetFinder({
    required this.basePath,
    required this.excludePaths,
  });

  /// Base path of the library.
  final String basePath;

  /// Paths which should be ignored. Usually something like `l10n/' to ignore
  /// the actual translation files.
  final List<ExcludePathChecker> excludePaths;

  final List<FoundRtlWidget> foundRtlWidgets = [];
  final Set<String> filesSkipped = <String>{};
  final Set<String> filesAnalyzed = <String>{};

  /// Starts the analyser and returns information about the found
  /// string literals.
  Future<List<FoundRtlWidget>> start() async {
    _logger.fine('Starting analysis.');
    final collection = AnalysisContextCollection(includedPaths: [basePath]);
    _logger.finer('Finding contexts.');
    for (final context in collection.contexts) {
      for (final filePath in context.contextRoot.analyzedFiles()) {
        final relative = path.relative(filePath, from: basePath);
        if (excludePaths.where((element) => element.shouldExclude(relative)).isNotEmpty ||
            // exclude generated code.
            filePath.endsWith('.g.dart')) {
          filesSkipped.add(filePath);
          continue;
        }
        filesAnalyzed.add(filePath);
        await _analyzeSingleFile(context, filePath);
      }
    }
    _logger.info('Found ${foundRtlWidgets.length} literals:');
    // for (final f in foundRtlWidgets) {
    //   final relative = path.relative(f.filePath, from: basePath);
    //   _logger.info('$relative:${f.loc} ${f.stringLiteral}');
    // }
    return foundRtlWidgets;
  }

  Future<void> _analyzeSingleFile(AnalysisContext context, String filePath) async {
    _logger.fine('analyzing $filePath');
//    final result = context.currentSession.getParsedUnit(filePath);
    final result = await context.currentSession.getResolvedUnit(filePath);
    if (result is! ResolvedUnitResult) {
      throw StateError('Did not resolve to valid unit.');
    }
    final unit = result.unit;
    final visitor = RtlWidgetVisitor<dynamic>(
        filePath: filePath,
        unit: unit,
        foundRtlWidget: (foundRtlWidget) {
          foundRtlWidgets.add(foundRtlWidget);
        });
    unit.visitChildren(visitor);
//    for (final unitMember in unit.declarations) {
//      _logger
//          .finest('${path.basename(filePath)} Found ${unitMember.runtimeType}');
//    }
  }
}

/// Information about a string literal found in dart code.
class FoundRtlWidget {
  FoundRtlWidget(
      {required this.filePath,
      required this.offset,
      required this.length,
      required this.msg,
      required this.error,
      required this.fix,
      this.startLine,
      this.startColumn,
      this.endLine,
      this.endColumn});

  /// absolute file path to the file in which the string literal was found.
  final String filePath;

  final int offset;

  final int length;

  /// The actual value of the string, better to use [stringLiteral].
  final String msg;

  final String error;

  final String fix;

  /// The one-based index of the line containing the first character of the
  /// range.
  int? startLine;

  /// The one-based index of the column containing the first character of the
  /// range.
  int? startColumn;

  /// The one-based index of the line containing the character immediately
  /// following the range.
  int? endLine;

  /// The one-based index of the column containing the character immediately
  /// following the range.
  int? endColumn;
}

class RtlWidgetVisitor<R> extends GeneralizingAstVisitor<R> {
  RtlWidgetVisitor({
    required this.filePath,
    required this.unit,
    required this.foundRtlWidget,
  }) : lineInfo = unit.lineInfo;

  static const loggerChecker = TypeChecker.fromRuntime(Logger);
  static const nonNlsChecker = TypeChecker.fromRuntime(NonNlsArg);
  static const exceptionChecker = TypeChecker.fromRuntime(Exception);
  static const errorChecker = TypeChecker.fromRuntime(Error);
  static const ignoredConstructorCalls = [
    TypeChecker.fromRuntime(Uri),
    TypeChecker.fromRuntime(RegExp),
    TypeChecker.fromUrl('package:flutter/src/painting/image_resolution.dart#AssetImage'),
    TypeChecker.fromUrl('package:flutter/src/widgets/navigator.dart#RouteSettings'),
    TypeChecker.fromUrl('package:flutter/src/foundation/key.dart#ValueKey'),
    TypeChecker.fromUrl('package:flutter/src/services/platform_channel.dart#MethodChannel'),
    TypeChecker.fromRuntime(StateError),
    loggerChecker,
    exceptionChecker,
    errorChecker,
  ];

  final String filePath;
  final CompilationUnit unit;
  final LineInfo? lineInfo;
  final void Function(FoundRtlWidget foundRtlWidget) foundRtlWidget;
  static final _evaluator = new ConstantEvaluator();

  @override
  R? visitInstanceCreationExpression(InstanceCreationExpression node) {
    if (_shouldIgnore(node)) {
      return null;
    }

    final constructorName = node.constructorName.toString();

    if (constructorName == 'Border.only' ||
        constructorName == 'BorderRadius.only' ||
        constructorName == 'BorderRadius.horizontal' ||
        constructorName == 'Positioned' ||
        constructorName == 'Positioned.fill' ||
        constructorName == 'EdgeInsets' ||
        constructorName == 'EdgeInsets.only' ||
        constructorName == 'EdgeInsets.fromLTRB') {
      bool shouldFix = false;
      List<String> arguments = [];

      node.argumentList.arguments.forEach((element) {
        if (element is! NamedExpression) return;

        NamedExpression express = element;
        var result = express.expression.accept(_evaluator);
        if (result == ConstantEvaluator.NOT_A_CONSTANT) {
          result = express.expression.toString();
        }

        String label = express.name.toString();
        if (label == 'left:') {
          label = 'start:';
          shouldFix = true;
        }
        if (label == 'right:') {
          label = 'end:';
          shouldFix = true;
        }
        String split = ' ';

        arguments.add("$label$split$result");
      });
      if (!shouldFix) return null;

      final constructorNameType = node.constructorName.type.toString();
      String fixResult = constructorName;

      fixResult = fixResult.replaceFirst(constructorNameType, "${constructorNameType}Directional");

      if ('Positioned.fill' == constructorName) fixResult = 'PositionedDirectional';
      if ('EdgeInsets.fromLTRB' == constructorName) fixResult = 'EdgeInsetsDirectional.fromSTEB';

      fixResult += '( ';
      fixResult += arguments.join(',');
      fixResult += ' )';

      final lineInfo = unit.lineInfo!;
      final begin = node.beginToken.charOffset;
      final end = node.endToken.charEnd;
      final loc = lineInfo.getLocation(begin);
      final locEnd = lineInfo.getLocation(end);

      foundRtlWidget(FoundRtlWidget(
          filePath: filePath,
          offset: node.offset,
          length: node.length,
          startLine: loc.lineNumber,
          startColumn: loc.columnNumber,
          endLine: locEnd.lineNumber,
          endColumn: locEnd.columnNumber,
          fix: fixResult,
          msg: 'RTL:Use ${constructorNameType}Directional instead',
          error: '$constructorName is not RTL campatiable'));
    }

    return super.visitInstanceCreationExpression(node);
  }

  @override
  R? visitPrefixedIdentifier(PrefixedIdentifier node) {
    if (_shouldIgnore(node)) {
      return null;
    }

    final prefix = node.prefix.name;
    if ('Alignment' == prefix) {
      bool shouldFix = false;
      String fixResult = node.toString().replaceFirst(prefix, "${prefix}Directional");
      if (node.identifier.name.contains('Left')) {
        fixResult = fixResult.replaceFirst('Left', 'Start');
        shouldFix = true;
      }

      if (node.identifier.name.contains('Right')) {
        fixResult = fixResult.replaceFirst('Right', 'End');
        shouldFix = true;
      }

      if (!shouldFix) return null;

      final lineInfo = unit.lineInfo!;
      final begin = node.beginToken.charOffset;
      final end = node.endToken.charEnd;
      final loc = lineInfo.getLocation(begin);
      final locEnd = lineInfo.getLocation(end);

      foundRtlWidget(FoundRtlWidget(
          filePath: filePath,
          offset: node.offset,
          length: node.length,
          startLine: loc.lineNumber,
          startColumn: loc.columnNumber,
          endLine: locEnd.lineNumber,
          endColumn: locEnd.columnNumber,
          msg: 'RTL:Use ${prefix}Directional instead',
          fix: fixResult,
          error: '$prefix is not RTL campatiable'));
    }

    return super.visitPrefixedIdentifier(node);
  }

  bool _checkArgumentAnnotation(
      ArgumentList argumentList, ExecutableElement? executableElement, Expression nodeChildChild) {
    final argPos = argumentList.arguments.indexOf(nodeChildChild);
    assert(argPos != -1);
    final arg = argumentList.arguments[argPos];
    ParameterElement param;
    if (arg is NamedExpression) {
      param =
          executableElement!.parameters.firstWhere((element) => element.isNamed && element.name == arg.name.label.name,
              orElse: () => throw StateError('Unable to find parameter of name ${arg.name.label} for '
                  '$executableElement'));
    } else {
      param = executableElement!.parameters[argPos];
      assert(param.isPositional);
    }
    if (nonNlsChecker.hasAnnotationOf(param)) {
//      _logger.finest('XX Argument is annotated with NonNls.');
      return true;
    }
    return false;
  }

  bool _shouldIgnore(AstNode origNode) {
    AstNode? node = origNode;
    AstNode? nodeChild;
    AstNode? nodeChildChild;
    for (; node != null; nodeChildChild = nodeChild, nodeChild = node, node = node.parent) {
      try {
        if (node is ImportDirective || node is PartDirective || node is PartOfDirective) {
          return true;
        }
        if (node is Annotation) {
          _logger.finest('Ignoring annotation parameters $node');
          return true;
        }
        if (node is ClassDeclaration) {
          if (nonNlsChecker.hasAnnotationOf(node.declaredElement!)) {
            if (nodeChild is FieldDeclaration) {
              if (nodeChild.isStatic) {
                return true;
              }
            }
          }
        }
        if (node is IndexExpression) {
          final target = node.realTarget;
          if (target is SimpleIdentifier) {
            try {
              if (nonNlsChecker.hasAnnotationOf(target.staticElement!)) {
                return true;
              }
            } catch (e, stackTrace) {
              _logger.warning('Unable to check annotation for $origNode at $filePath', e, stackTrace);
            }
          }
        }
        if (node is InstanceCreationExpression) {
          assert(nodeChild == node.argumentList);
          if (_checkArgumentAnnotation(
              node.argumentList, node.constructorName.staticElement, nodeChildChild as Expression)) {
            return true;
          }
//        param.no
//          node.constructorName.staticElement;
          for (final ignoredConstructorCall in ignoredConstructorCalls) {
            if (ignoredConstructorCall.isAssignableFrom(node.staticType!.element!)) {
              return true;
            }
          }
        }
        if (node is VariableDeclaration) {
          final element = node.declaredElement;
          if (element != null && nonNlsChecker.hasAnnotationOf(element)) {
            return true;
          }
        }
        if (node is FormalParameter) {
          final element = node.declaredElement;
          if (element != null && nonNlsChecker.hasAnnotationOf(element)) {
            return true;
          }
        }
        if (node is MethodInvocation) {
          if (nodeChildChild is! Expression) {
            _logger.warning('not an expression. $nodeChildChild ($node)');
            // } else if (nodeChildChild != origNode) {
            //   we only care about direct method calls.
          } else if (
              // check if `nodeChildChild` is actually a full argument.
              // this can happen with sub expressions like
              // myFunc('string'.split('').join('')); where
              // `string'.split('')` will not be found in the parent expression.
              node.argumentList.arguments.contains(nodeChildChild) &&
                  // check if the argument is annotated
                  _checkArgumentAnnotation(
                      node.argumentList, node.methodName.staticElement as ExecutableElement?, nodeChildChild)) {
            return true;
          }
          final target = node.target;
          if (target != null) {
            // ignore all calls to `Logger`
            if (target.staticType == null) {
              _logger.warning('Unable to resolve type for $target');
            } else if (loggerChecker.isAssignableFromType(target.staticType!)) {
              return true;
            }
          }
        }
        if (node is FunctionDeclaration || node is MethodDeclaration) {
          if (node is Declaration) {
            if (nonNlsChecker.hasAnnotationOf(node.declaredElement!)) {
              return true;
            }
          }
        }
      } catch (e, stackTrace) {
        final loc = lineInfo!.getLocation(origNode.offset);
        _logger.severe('Error while analysing node $origNode at $filePath $loc', e, stackTrace);
      }
    }
    // see if we can find a line end comment.
    final lineNumber = lineInfo!.getLocation(origNode.end).lineNumber;
    var nextToken = origNode.endToken.next;
    while (nextToken != null && lineInfo!.getLocation(nextToken.offset).lineNumber == lineNumber) {
      nextToken = nextToken.next;
    }
    final comment = nextToken!.precedingComments;
    if (comment != null && lineInfo!.getLocation(comment.offset).lineNumber == lineNumber) {
      if (comment.value().contains('NON-NLS')) {
        return true;
      }
    }
    return false;
  }
}
